/****************************************************************************
**
** Copyright (C) 2006-2009 fullmetalcoder <fullmetalcoder@hotmail.fr>
**
** This file is part of the Edyuk project <http://edyuk.org>
** 
** This file may be used under the terms of the GNU General Public License
** version 3 as published by the Free Software Foundation and appearing in the
** file GPL.txt included in the packaging of this file.
**
** This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
**
****************************************************************************/

#ifndef _QDOCUMENT_P_H_
#define _QDOCUMENT_P_H_

#include "qce-config.h"

/*!
	\file qdocument_p.h
	\brief Definition of the private document API
*/

#include "qdocument.h"
#include "qdocumentline.h"
#include "qdocumentcursor.h"

#include <QHash>
#include <QFont>
#include <QStack>
#include <QQueue>
#include <QDateTime>
#include <QUndoStack>
#include <QStringList>
#include <QFontMetrics>
#include <QUndoCommand>

class QEditConfig;
class QMarksManager;

class QDocument;
class QDocumentMark;
class QDocumentBuffer;
class QDocumentPrivate;
class QDocumentCommand;
class QDocumentRegionTree;
class QDocumentCommandBlock;

class QLanguageDefinition;

Q_DECLARE_TYPEINFO(QDocumentSelection, Q_PRIMITIVE_TYPE);

#include "qdocumentline_p.h"

#include "qdocumentcursor_p.h"

#define _USE_DOC_BUFFER_

#include "qdocumentbuffer.h"

class QCE_EXPORT QDocumentPrivate
{
	friend class QEditConfig;
	friend class QMarksManager;
	
	friend class QDocument;
	friend class QDocumentCommand;
	friend class QDocumentRegionTree;
	friend class QDocumentLineHandle;
	friend class QDocumentCursorHandle;
	
	public:
		QDocumentPrivate(QDocument *d);
		~QDocumentPrivate();
		
		void execute(QDocumentCommand *cmd);
		
		void draw(QPainter *p, QDocument::PaintContext& cxt);
		
		QDocumentLineHandle* at(int line) const;
		int indexOf(const QDocumentLineHandle *l) const;
		
		int wrappedLineForCursor(QDocumentLineHandle *h, int cpos);
		
		QPoint cursorToDocumentOffset(QDocumentLineHandle *h, int cpos);
		
		int documentOffsetToCursor(QDocumentLineHandle *h, int x, int y);
		void cursorToDocumentOffset(QDocumentLineHandle *h, int cpos, int& x, int& y);
		
		QVector<int> wrapBoundaries(QDocumentLineHandle *h) const;
		
		bool updateWrapBoundaries(QDocumentLineHandle *h, int *span = 0);
		void computeWrapBoundaries(QDocumentLineHandle *h, QVector<int>& wb);
		
		void compose(QDocumentLineHandle *h);
		void layout(QDocumentLineHandle *h);
		
		void draw(
					QDocumentLineHandle *h,
					QPainter *p,
					int xOffset,
					int vWidth,
					const QVector<int>& sel,
					const QVector<int>& cursors,
					const QPalette& pal,
					bool fullSel
				);
		
		void adjustWidth(int l);
		
		void setWidth();
		void setHeight();
		
		static void setFont(const QFont& f);
		
		void beginChangeBlock();
		void endChangeBlock();
		
		inline int maxMarksPerLine() const
		{ return m_maxMarksPerLine; }
		
		inline bool hasMarks() const
		{ return m_marks.count(); }
		
		QList<int> marks(QDocumentLineHandle *h) const;
		
		void addMark(QDocumentMark *m);
		void removeMark(QDocumentMark *m);
		
		void addMark(int line, int mid);
		void toggleMark(int line, int mid);
		void removeMark(int line, int mid);
		
		void addMark(QDocumentLineHandle *h, int mid);
		void toggleMark(QDocumentLineHandle *h, int mid);
		void removeMark(QDocumentLineHandle *h, int mid);
		
		int findNextMark(int id, int from = 0, int until = -1);
		int findPreviousMark(int id, int from = -1, int until = 0);
		
		QList<QDocumentMark*> findMarks(QDocumentCursorHandle *h, const QList<int>& ids) const;
		
		int getNextGroupId();
		void releaseGroupId(int groupId);
		void clearMatches(int gid);
		void flushMatches(int gid);
		void addMatch(int gid, int line, int pos, int len, int format);
		
		void emitFormatsChange (int line, int lines);
		void emitContentsChange(int line, int lines);
		
		int visualLine(int textLine) const;
		int textLine(int visualLine, int *wrap = 0) const;
		void hideEvent(int line, int count);
		void showEvent(int line, int count);
		
		void setWidth(int width);
		
		void emitFormatsChanged();
		void emitContentsChanged();
		
		void emitLineDeleted(QDocumentLineHandle *h);
		
		void emitMarkAdded(QDocumentMark *m);
		void emitMarkRemoved(QDocumentMark *m);
		void emitMarkChanged(QDocumentMark *m);
		
		//void emitMarkChanged(QDocumentLineHandle *l, int m, bool on);
		
		inline int lineCount() const
		{
			return m_buffer->count();
		}
		
		QDocumentBuffer::iterator begin() const;
		QDocumentBuffer::iterator end() const;
		
		QDocumentBuffer::iterator find(int line) const;
		
	protected:
		void insertLines(int after, const QVector<QDocumentLineHandle*>& l);
		void removeLines(int after, int n);
		
		void emitWidthChanged();
		void emitHeightChanged();
		
		void updateFormatCache();
		void setFormatScheme(QFormatScheme *f);
		void tunePainter(QPainter *p, int fid);
		
	private:
		QDocument *m_doc;
		QUndoStack m_commands;
		QDocumentCursor *m_editCursor;
		
		bool m_suspend, m_deleting;
		QQueue<QPair<int, int> > m_notifications;
		
		QHash<QDocumentLineHandle*, QVector<int> > m_boundaries;
		QVector< QPair<QDocumentLineHandle*, int> > m_largest;
		
		struct Match
		{
			int line;
			QFormatRange range;
			QDocumentLineHandle *h;
		};
		
		struct MatchList : QList<Match>
		{
			MatchList() : index(0) {}
			
			int index;
		};
		
		int m_lastGroupId;
		QList<int> m_freeGroupIds;
		QHash<int, MatchList> m_matches;
		
		bool m_constrained;
		int m_width, m_height;
		
		int m_tabStop;
		static int m_defaultTabStop;
		
		static QFont *m_font;
		static bool m_fixedPitch;
		static QFontMetrics *m_fontMetrics;
		static int m_leftMargin;
		static QDocument::WhiteSpaceMode m_showSpaces;
		static QDocument::LineEnding m_defaultLineEnding;
		static int m_lineHeight;
		static int m_lineSpacing;
		static int m_spaceWidth;
		static int m_ascent;
		static int m_descent;
		static int m_leading;
		static int m_wrapMargin;
		
		QFormatScheme *m_formatScheme;
		QLanguageDefinition *m_language;
		static QFormatScheme *m_defaultFormatScheme;
		
		QVector<QFont> m_fonts;
		
		static QList<QDocumentPrivate*> m_documents;
		
		int m_maxMarksPerLine;
		QList<QDocumentMark*> m_marks;
		QHash<QDocumentLineHandle*, QList<QDocumentMark*> > m_marksLookup;
		
		QHash<QDocumentLineHandle*, QPair<int, int> > m_status;
		
		QString m_lineEndingString;
		QDocument::LineEnding m_lineEnding;
		
		QDateTime m_lastModified;
		QDocumentBuffer *m_buffer;
		QDocumentRegionTree *m_regions;
};

#endif
