/****************************************************************************
**
** Copyright (C) 2006-2009 fullmetalcoder <fullmetalcoder@hotmail.fr>
**
** This file is part of the Edyuk project <http://edyuk.org>
** 
** This file may be used under the terms of the GNU General Public License
** version 3 as published by the Free Software Foundation and appearing in the
** file GPL.txt included in the packaging of this file.
**
** This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
**
****************************************************************************/

#include "qnfadefinition.h"

/*!
	\file qnfadefinition.cpp
	\brief Implementation of the QNFADefinition class.
*/

#include "qnfa.h"

#include "qformatscheme.h"

#include "qmarksmanager.h"

#include "qdocument.h"
#include "qdocument_p.h"
#include "qdocumentline.h"
#include "qdocumentcursor.h"

#include <QFile>
#include <QKeyEvent>

#include <QDebug>
#include <QDomText>
#include <QDomElement>
#include <QDomDocument>

uint qHash(const QPointer<QDocument>& p) { return qHash((QDocument*)(p)); }

extern QString *_singleLineCommentTarget;
void embed(QNFA *src, QNFA *dest, int index);
void fillContext(QNFA *cxt, QVector<QNFA*>& cxtIds, QDomElement e, QFormatScheme *f,
				QVector<QNFADefinition::Action*>& actions,
				QVector<QNFADefinition::ParenGroup*>& parens, bool cs);

extern bool stringToBool(const QString& s, bool previous);

QHash<QString, QNFA*> QNFADefinition::m_contexts;
QHash<QString, QNFADefinition::EmbedRequestList> QNFADefinition::m_pendingEmbeds;

void QNFADefinition::load(const QString& file, QLanguageFactory::LangData *d, QFormatScheme *s)
{
	QFile f(file);
	
	if ( !f.open(QFile::ReadOnly | QFile::Text) )
	{
		qWarning("QNFADefinition : failed to open file %s", qPrintable(file));
		return;
	}
	
	load(&f, d, s);
}

void QNFADefinition::load(QFile *f, QLanguageFactory::LangData *d, QFormatScheme *s)
{
	QDomDocument doc;
	doc.setContent(f);
	
	load(doc, d, s);
}

void QNFADefinition::load(const QDomDocument& doc, QLanguageFactory::LangData *d, QFormatScheme *s)
{
	QDomElement root = doc.documentElement();
	
	QNFADefinition *nd = new QNFADefinition;
	
	nd->m_language = d->lang = root.attribute("language");
	nd->m_indentFold = stringToBool(root.attribute("indentationFold"), false);
	nd->m_extensions = d->extensions = root.attribute("extensions").split(";");
	nd->m_defaultMark = root.attribute("defaultLineMark", "bookmark");
	
	/*
	qDebug("Generating definition for %s language from XML file : %s",
			qPrintable(m_language),
			qPrintable(fn));
	*/
	
	// create root entity
	nd->m_root = lexer();
	nd->m_contextIds << nd->m_root;
	
	_singleLineCommentTarget = &(nd->m_singleLineComment);
	fillContext(nd->m_root, nd->m_contextIds, root, s, nd->m_actions, nd->m_parens, true);
	_singleLineCommentTarget = 0;
	
	//qDebug("%s : %i, %i", qPrintable(nd->m_language), nd->m_actions.count(), nd->m_parens.count());
	
	squeeze(nd->m_root);
	
	//m_definitions[m_language] = this;
	m_contexts[nd->m_language] = nd->m_root;
	
	flushEmbedRequests(nd->m_language);
	
	d->d = nd;
	d->e = 0;
	d->s = s;
}

void QNFADefinition::registerContext(QNFA *a)
{
	
}

QNFADefinition::QNFADefinition()
 : m_indentFold(false), m_root(0)
{
	
}

QNFADefinition::~QNFADefinition()
{
	delete m_root;
}

QString QNFADefinition::language() const
{
	return m_language;
}

QStringList QNFADefinition::extensions() const
{
	return m_extensions;
}

inline bool isWord(QChar c)
{ return c.isLetterOrNumber() || (c.unicode() == '_'); }

static inline bool match(QChar cc, const QNFA *chain)
{
	bool found = !(n && !fc);
	const quint16 cu = cc.unicode();
	const int n = chain->c.count();
	const int ass = chain->assertion;
	quint16 fc = n ? chain->c.at(0) : 0;
	
	// avoid conditionals at all cost...
	
	const bool fallback = chain->c.contains(cu) ? found : !found;
	const bool isDigit = cc.isDigit();
	const bool isSpace = cc.isSpace();
	const bool isWord = cc.isLetterOrNumber() || cu == '_';
	
	return
			((n == 1) && (fc == cu))
		?
			true
		:
			(
				ass
			?
				(
					(
						((ass & Digit) && isDigit)
					||
						((ass & Space) && isSpace)
					||
						((ass & Word) && isWord)
					||
						((ass & NonDigit) && !isDigit)
					||
						((ass & NonSpace) && !isSpace)
					||
						((ass & NonWord) && !isWord)
					)
				?
					found
				:
					fallback
				)
			:
				fallback
			)
		;
	
	/*
	if ( chain->c.count() == 1 )
	{
		if ( fc == cu )
			return found;
	}
	
	if ( notEmpty && (fc == '\0') )
		found = false;
	
	
	
	if ( ass )
	{
		if ( cc.isDigit() )
		{
			if ( ass & Digit )
				return found;
		} else {
			if ( ass & NonDigit )
				return found;
			
			if ( cc.isSpace() )
			{
				if ( ass & Space )
					return found;
			} else {
				if ( ass & NonSpace )
					return found;
				
				if ( cc.isLetterOrNumber() || (cu == '_') )
				{
					if ( ass & Word )
						return found;
				} else {
					if ( ass & NonWord )
						return found;
				}
			}
		}
	}
	
	if ( notEmpty )
	{
		if ( chain->c.contains(cu) )
			return found;
	}
	
	return !found;
	*/
}

void processLine()
{
	
}

/*!
	\brief Entry point for syntax highlighting
*/
int QNFADefinition::tokenize(
								QDocumentBuffer::iterator& it,
								int& m_line,
								int& m_column,
								int extent,
								int depth,
								int contCxtIdx)
{
	QNFA *context;
	QStack<QNFA*> cxt;
	QVector<int> minimal;
	bool cameBack = false;
	const int startLine = m_line, lastLine = extent;
	
	cxt.append(m_root);
	
	if ( m_line > 0 && (!depth || contCxtIdx != -1) )
	{
		int min = 0;
		QDocumentBuffer::iterator tmp = it;
		tmp.moveBy(-1);
		
		const QVector<int>& cxts = tmp.lineHandle()->contexts();
		
		if ( depth )
		{
			min = contCxtIdx;
			
			/*
			// build up context stack to bring along in case of line span...
			
			for ( int i = 0; i + 2 < min; i += 3 )
				minimal << cxts[i] << cxts[i + 1] << cxts[i + 2];
			*/
		}
		
		for ( int i = min; i + 2 < cxts.count(); i += 3 )
		{
			if ( cxts.at(i + 1) != -1 )
				break;
			
			// at this point we're only interested in context going way back
			// (for a certain value of way back, i.e at list one line...)
			
			int cxtId = cxts.at(i + 2);
			
			if ( cxtId & Language )
			{
				// switching language... must forward the thing...
				QLanguageDefinition *d = QLanguageDefinition::instance(cxt & IdMask);
				
				if ( d )
				{
					int covered = d->tokenize(it, m_line, m_column, extent, depth + 1, i + 3);
					
					if ( covered != -1 )
					{
						// did something, we're done with the tricky continuation part
						
						// now onto the tricky exit part (or trickier even continuation?)
						cameBack = true;
						
						if (
								m_line > lastLine
							||
								(m_line == lastLine && m_column == it.lineHandle()->length())
							)
						{
							// went beyond minimal scope
							
						} else {
							// now we can just happily continue our road with the context stack
							// we've found so far from the point the embedded language reached.
							
							break;
						}
					} else {
						// ... what the heck could we do???
						qWarning("broken continuation of embedded language");
						return extent - m_line;
					}
				} else {
					qWarning("Ought to embed but unable to retrieve language definition...");
					return extent - m_line;
				}
			} else {
				cxt.append(m_contextIds[cxtId]);
			}
		}
	}
	
	// pop the current context
	context = cxt.pop();
	
	int offset, idx, length;
	QDocumentLineHandle *h = 0, *prev;
	
	// lexing loop, line by line
	while ( !it.atEnd() )
	{
		prev = h;
		h = it.lineHandle();
		
		if ( m_column >= h->length() )
		{
			++m_line;
			m_column = 0;
			it.moveBy(1);
			continue;
		}
		
		const QChar *d = h->text().constData();
		
		idx = m_column;
		length = h->length();
		
		QVector<int> oldCxtStack, newCxtStack;
		
		if ( (m_line + 1 >= lastLine) && (m_line > 0) )
		{
			if ( !prev )
			{
				QDocumentBuffer::iterator tmp = it;
				tmp.moveBy(-1);
				prev = tmp.lineHandle();
			}
			
			const QVector<int>& cxts = prev->contexts();
			
			for ( int i = 0; i + 2 < cxts.count(); i += 3 )
			{
				if ( cxts.at(i + 1) == -1 )
				{
					int cxtId = cxts.at(i + 2);
					
					oldCxtStack.append(cxtId);
				}
			}
		}
		
		if ( (!depth || (m_line != startLine)) && !cameBack )
		{
			h->clearTokens();
			h->clearContexts();
		}
		
		cameBack = false;
		
		QStack<int> cxtidx;
		QMap<int, QPair<int, int> > cxtData;
		
		QList<QLanguageDefinition*> embeds = m_embeds.value(context);
		
		bool beRightBack = false;
		
		{
			quint16 c = 0;
			const QChar *di = d;
			QNFA *chain = 0, *start = 0;
			int index = idx, lastCxt = 0, len;
			bool bFound, bEscape = false, bEscaped = false;
			bool wPrev = false, wCur = false;
			
			while ( index <= length )
			{
				bFound = false;
				bEscaped = false;
				
				if ( index == length )
				{
					c = '\n';
					wCur = false;
				} else {
					c = di->unicode();
					wCur = isWord(*di);
				}
				
				int embedIdx = 0;
				int plainIndex = -1, plainMatch, plainLength;
				
				// pre-embeds
				while ( embedIdx < embeds.count() )
				{
					QLanguageDefinition *embedded = embeds.at(embedIdx);
					
					if ( !embedded )
						break;
					
					int n = embedded->tokenize(it, m_line, m_column, extent, depth + 1, -1);
					
					if ( n == -1 )
					{
						++embedIdx;
					} else if ( !n ) {
						// something short matched
						beRightBack = true;
						
						
					} else {
						// something long matched
						cameBack = true;
						break;
					}
				}
				
				if ( beRightBack )
				{
					continue;
				}
				
				if ( cameBack )
				{
					break;
				}
				
				// try fast plain matching
				if ( !(wPrev && wCur) )
				{
					//qDebug("trying plain...");
					
					//len = 0;
					idx = index;
					QCharTree::const_iterator it, match, end;
					
					it = context->tree.constFind(c);
					
					if ( it != context->tree.constEnd() )
					{
						//qDebug("plain on %c", c);
						do
						{
							++di;
							++idx;
							
							end = it->next.constEnd();
							match = it->next.constFind(0);
							
							if ( idx <= length )
							{
								c = idx == length ? '\n' : di->unicode();
								it = it->next.constFind(c);
							} else {
								it = end;
							}
							
							if ( it == end )
							{
								if ( (match != end) && !isWord(*di) )
								{
									//word boundary found
									// corresponding token end found
									wPrev = isWord(*(di - 1));
									bFound = true;
									if ( match->value.action & 0x40000000 )
									{
										// try regexps before notifying
										plainIndex = index;
										plainLength = idx - index;
										plainMatch = match->value.action;
										//qDebug("ambiguity.");
									} else {
										h->addToken(index, idx - index, match->value.action);
										index = idx;
									}
									//qDebug("next step : %c", d[index].toLatin1());
									//bMonitor = true;
								}
								
								break;
							}
						} while ( idx < length );
						
						if ( bFound )
						{
							bEscape = false;
							
							if ( plainIndex == -1 )
								continue;
							
							bFound = false;
						}
						
						di -= idx - index;
					}
				}
				
				// fallback on regexp-like NFA-based matching
				QNFABranch* children = context->out.branch;
				
				if ( children )
				{
					//qDebug("trying %i sub nfas on %c", children->count(), d[index].toLatin1());
					const int max = children->count();
					
					for ( quint16 i = 0; i < max; ++i )
					{
						len = 0;
						idx = index;
						start = chain = children->at(i);
						
						//qDebug("%ith attempt on %c", i, d[index + len].toLatin1());
						
						while ( (idx <= length) || (chain->type & Match) )
						{
							bEscaped = false;
							QChar cc = idx < length ? *di : '\n';
							
							if ( chain->type & Match )
							{
								if (
										(chain->assertion & WordEnd)
									&&
										isWord(*di)
									&&
										isWord(*(di - 1))
									)
								{
									//qDebug("end assertion failed...");
									break;
								}
								
								//qDebug("matched to end");
								
								if ( chain->type & CxtBeg )
								{
									//qDebug("entering context : 0x%x", chain);
									
									// TODO : enter embedded
									
									h->addToken(index, len, start->actionid ? start->actionid : chain->actionid);
									
									// update context stack
									cxt.push(context);
									
									context = chain;
									cxtidx.push(idx);
									//cxtid.push(m_contextIds.indexOf(context)); //chain->actionid);
									//qDebug("entering %i (0x%x)", cxtid.top(), context);
									
									getEmbedList(context, pre_embeds, post_embeds);
									
								} else if ( chain->type & CxtEnd ) {
									//qDebug("leaving context :");
									
									if ( cxt.isEmpty() )
										qFatal("context nesting problem");
									
									if ( bEscape )
									{
										// not really end : escape found...
										
										bEscape = false;
										bEscaped = true;
									} else {
										int cxtlen;
										
										if ( cxtidx.count() )
										{
											lastCxt = cxtidx.pop();
											cxtlen = index - lastCxt;
										} else {
											lastCxt = -1;
											cxtlen = index;
										}
										
										h->addToken(index, len, chain->actionid);
										
										//qDebug("[%i, %i] : %i", lastCxt, len, context->actionid);
										//h->addContext(lastCxt, cxtlen, context->actionid);
										
										if ( lastCxt != -1 || !(context->type & StayOnLine) )
											cxtData.insertMulti(lastCxt, qMakePair(cxtlen, m_contextIds.indexOf(context))); //context->actionid);
										
										// update context stack
										context = cxt.pop();
										
										getEmbedList(context, pre_embeds, post_embeds);
										
										if ( cxtidx.isEmpty() )
										{
											while ( cxt.count() && (context->type & StayOnLine) )
											{
												//qDebug("down");
												context = cxt.pop();
											}
										}
										
										//qDebug("cxt notif...");
										
										if ( chain->type & Exclusive )
										{
											index = idx;
										} else {
											//qDebug("non-exclusive");
										}
										
										--index;
										--di;
										
										bFound = true;
										break;
									}
								} else if ( chain->type & CxtEsc )  {
									//qDebug("matched %s", qPrintable(QString(index, len)));
									
									//notify(index, len, chain->actionid);
									bEscape = !bEscape;
								} else {
									//qDebug("matched %s", qPrintable(QString(d + index, len)));
									
									if ( plainIndex != -1 && plainLength >= len )
									{
										break;
									}
									
									//notify(index, len, chain->actionid);
									h->addToken(index, len, chain->actionid);
									bEscape = false;
								}
								
								bFound = true;
								index = idx;
								--index;
								--di;
								
								//qDebug("next step : %c", d[index + 1].toLatin1());
								//bMonitor = true;
								
								break;
							} else {
								// "regular" nfa match (no match yet...)
								
								if (
										(chain->assertion & WordStart)
									&&
										(idx >= 1)
									&&
										(
											isWord(*(di - 1))
										&&
											isWord(cc)
										)
									)
								{
									//qDebug("beg assertion failed...");
									
									break;
								}
								
								bool found = ::match(cc, chain);
								
								if (
										!(chain->assertion & ZeroOrOne)
									&&
										!(chain->assertion & ZeroOrMore)
									&&
										!found
									)
								{
									//if ( cc.toLatin1() == ')' )
									//	qDebug("mismatch : %c != %c", cc.toLatin1(), chain->c.at(0));
									
									break;
								}
								
								if ( found )
								{
									//qDebug("%c", d[index + len].toLatin1());
									
									if (
											(chain->assertion & OneOrMore)
										||
											(chain->assertion & ZeroOrMore)
										)
									{
										do
										{
											++di;
											++len;
											++idx;
											
											if ( idx > length )
												break;
											
											if ( idx == length )
												cc = eol;
											else
												cc = *di;
											
										} while ( ::match(cc, chain) );
										
									} else {
										++len;
										++idx;
										++di;
									}
									
								} else {
									//qDebug("! %c", d[index + len].toLatin1());
								}
								
								chain = chain->out.next;
							}
						}
						
						if ( bFound )
							break;
						
						di -= len;
					}
				}
				
				if ( !bFound )
				{
					if ( plainIndex != -1 )
					{
						//notify(plainIndex, plainLength, plainMatch);
						h->addToken(plainIndex, plainLength, plainMatch);
						index = plainIndex + plainLength;
						di += plainLength;
						continue;
					}
					
					if ( depth && !inside )
					{
						// submatch failed, go up
						return 0;
						//return m_line - line;
					}
					
					bEscape = false;
					//++index;
					wPrev = wCur;
				} else if ( index < length ) {
					wPrev = isWord(*di);
				}
				
				++index;
				++di;
			}
			
			if ( cameBack )
				continue;
		}
		
		++m_line;
		// commit contexts in proper order
		int n = cxt.count(), nn = cxtidx.count();
		
		int anchor = nn ? cxtidx.last() : -1;
		
		if ( nn )
			--nn;
		
		int i = 1;
		
		//qDebug("n:%i, nn:%i", n, nn);
		
		while ( i < (n - nn) )
		{
			int cxtId = m_contextIds.indexOf(cxt.at(i)); //cxtid.at(i);//m_contextIds.indexOf(cxt.at(i))  | languageId();
			
			if ( cxt.at(i)->type & StayOnLine )
			{
				//qDebug("oldstay");
				--n;
				cxt.remove(i);
				continue;
			}
			
			//qDebug(">][ : %i", cxtId);
			h->addContext(-1, -1, cxtId);
			
			++i;
			
			if ( m_line >= lastLine )
			{
				newCxtStack.append(cxtId);
			}
		}
		
		int j = i;
		
		if ( !(n - nn) )
		{
			i = 0;
		}
		
		while ( j < n )
		{
			int k = j - i;
			
			int cxtId = m_contextIds.indexOf(cxt.at(j)); //cxtid.at(k);//m_contextIds.indexOf(cxt.at(i))  | languageId();
			cxtData.insertMulti(cxtidx.at(k), qMakePair(-1, cxtId));
			
			//qDebug(">%i[ : %i", cxtidx.at(k), cxtId);
			
			if ( cxt.at(j)->type & StayOnLine )
			{
				//qDebug("stay");
				--n;
				cxt.remove(j);
				cxtidx.remove(k);
				continue;
			}
			
			//qDebug(">[%i, [ : %i", cxtidx.at(j - i), cxtId);
			++j;
		}
		
		//newCxtStack = cxtid;
		
		if ( cxt.count() )
		{
			int cxtId = m_contextIds.indexOf(context); // | languageId();
			
			if ( anchor != -1 )
			{
				cxtData.insertMulti(anchor, qMakePair(-1, cxtId));
			} else {
				//qDebug(">][ : %i (top)", cxtId);
				
				if ( m_line >= lastLine )
					newCxtStack.append(cxtId);
				
				h->addContext(-1, -1, cxtId);
			}
		}
		
		QMap<int, QPair<int, int> >::const_iterator cit = cxtData.constBegin();
		
		while ( cit != cxtData.constEnd() )
		{
			//qDebug(">[%i,%i? : %i", cit.key(), cit->first, cit->second);
			h->addContext(cit.key(), cit->first, cit->second);
			
			if ( cit->first == -1 && (m_line >= lastLine) )
			{
				newCxtStack.append(cit->second);
			}
			
			++cit;
		}
		
		cxtidx.clear();
		it.moveBy(1);
		m_column = 0;
		
		//qDebug() << cxt << context;
		
		if ( m_line >= lastLine )
		{
			// check whether further lexing is required
			//qWarning("TODO : continuation");
			
			//qDebug() << oldCxtStack << " vs " << newCxtStack << endl;
			
			if ( newCxtStack == oldCxtStack )
				break;
			
			//qDebug("continuing");
			newCxtStack.clear();
			oldCxtStack.clear();
			//break;
		}
	}
	
	return m_line - startLine;
}

/*!
	\brief Return the string starting a single line comment, if any offered by the language
*/
QString QNFADefinition::singleLineComment() const
{
	return m_singleLineComment;
}

QString QNFADefinition::defaultLineMark() const
{
	return m_defaultMark;
}

/*!
	\brief Brace matching entry point
*/
void QNFADefinition::clearMatches(QDocument *d)
{
	QHash<QPointer<QDocument>, int>::iterator it = m_matchGroups.find(d);
	
	if ( it != m_matchGroups.end() )
	{
		d->clearMatches(*it);
		d->flushMatches(*it);
		
		// erase?
		//m_matchGroups.erase(it);
	}
}

/*!
	\brief Brace matching entry point
*/
void QNFADefinition::match(QDocumentCursor& c)
{
	QDocumentLine b = c.line();
	QDocument *d = c.document();
	
	if ( !d || !b.isValid() )
	{
		qDebug("invalid cursor");
		return;
	}
	
	int gid = 0;
	QHash<QPointer<QDocument>, int>::iterator it = m_matchGroups.find(d);
	
	if ( it != m_matchGroups.end() )
	{
		d->clearMatches(*it);
		d->flushMatches(*it);
		*it = gid = d->getNextGroupId();
	} else {
		gid = d->getNextGroupId();
		m_matchGroups[d] = gid;
	}
	
	QFormatScheme *s = d->formatScheme();
	
	if ( !s )
		s = QDocument::defaultFormatScheme();
	
	if ( !s )
	{
		qDebug("no fmt");
		d->releaseGroupId(gid);
		m_matchGroups.remove(d);
		return;
	}
	
	const int matchFID = s->id("braceMatch"), mismatchFID = s->id("braceMismatch");
	
	int matchCount = 0;
	int pos = c.columnNumber();
	
	const QVector<int>& m_tokens = b.tokens();
	
	if ( m_tokens.count() < 4 )
	{
		// required to properly update display
		d->releaseGroupId(gid);
		m_matchGroups.remove(d);
		return;
	}
	
	//qDebug("matching on line %i (fid is %i)", c.lineNumber(), fid);
	
	int tokenIndex = 1;
	int tokenCount = m_tokens.first();
	
	while ( tokenIndex + 2 < tokenCount )
	{
		int offset = m_tokens.at(tokenIndex);
		int length = m_tokens.at(++tokenIndex);
		int tokenId = normalized(m_tokens.at(++tokenIndex));
		++tokenIndex;
		
		if ( (pos != offset) && (pos != (offset + length)) )
			continue;
		
		if ( !(tokenId & QLanguageDefinition::BraceMatch) )
			continue;
		
		PMatch m;
		m.line[0] = c.lineNumber();
		m.column[0] = offset;
		m.length[0] = length;
		
		if ( (tokenId & QLanguageDefinition::OpenParen) && (tokenId & QLanguageDefinition::CloseParen) )
		{
			//qDebug("match bound");
			
			matchOpen(d, m);
			
			if ( m.type == PMatch::Match )
			{
				++matchCount;
				
				d->addMatch(gid, m.line[0], m.column[0], m.length[0], matchFID);
				d->addMatch(gid, m.line[1], m.column[1], m.length[1], matchFID);
			}
			
			m.line[0] = c.lineNumber();
			m.column[0] = offset;
			m.length[0] = length;

			matchClose(d, m);
			
			if ( m.type == PMatch::Match )
			{
				++matchCount;
				
				d->addMatch(gid, m.line[0], m.column[0], m.length[0], matchFID);
				d->addMatch(gid, m.line[1], m.column[1], m.length[1], matchFID);
			}
			
			continue;
		} else if ( tokenId & QLanguageDefinition::OpenParen ) {
			//qDebug("match open");
			
			matchOpen(d, m);
		} else if ( tokenId & QLanguageDefinition::CloseParen ) {
			//qDebug("match close");
			
			matchClose(d, m);
		}
		
		if ( m.type != PMatch::Invalid )
		{
			++matchCount;
			
			int mfid = m.type == PMatch::Match ? matchFID : mismatchFID;
			d->addMatch(gid, m.line[0], m.column[0], m.length[0], mfid);
			d->addMatch(gid, m.line[1], m.column[1], m.length[1], mfid);
		}
	}
	
	if ( matchCount )
	{
		d->flushMatches(gid);
	} else {
		//qDebug("match failed");
		
		d->releaseGroupId(gid);
		m_matchGroups.remove(d);
	}
}

int QNFADefinition::parenId(quint32 tokenId) const
{
	for ( int i; i < m_parens.count(); ++i )
	{
		if ( m_parens.at(i)->ids.contains(tokenId) )
			return i;
		
	}
	
	return -1;
}

void QNFADefinition::matchOpen(QDocument *d, PMatch& m)
{
	int line = m.line[0];
	int pos = m.column[0];
	QVector<int> m_tokens = d->line(line).tokens();
	
	bool bMatch = false;
	int offset, length;
	QStack<int> lines;
	QStack<bool> type;
	QStack<int> parens;
	QStack<int> columns;
	QStack<int> lengths;
	
	forever
	{
		if ( m_tokens.count() > 3 )
		{
		int tokenIndex = 1;
		int tokenCount = m_tokens.first();
		
		while ( tokenIndex + 2 < tokenCount )
		{
			offset = m_tokens.at(tokenIndex);
			length = m_tokens.at(++tokenIndex);
			int rawId = m_tokens.at(++tokenIndex);
			++tokenIndex;
			
			int tokenId = normalized(rawId);
			
			bool open = tokenId & QLanguageDefinition::OpenParen;
			bool close = tokenId & QLanguageDefinition::CloseParen;
			
			if ( !(open || close) || offset < pos )
				continue;
			
			int parenId = this->parenId(rawId);
			
			if ( open && close )
			{
				bool bInsert = true;
				
				if ( parens.count() )
				{
					if ( parens.top() == parenId )
					{
						//qDebug("submatch at %i", line);
						type.pop();
						lines.pop();
						parens.pop();
						columns.pop();
						lengths.pop();
						
						if ( parens.isEmpty() )
						{
							bMatch = true;
							//qDebug("match at %i", line);
							m.type = PMatch::Match;
							
							break;
						}
					} else {
						bInsert = false;
						
						do
						{
							int tp = parens.top();
							
							if ( type.top() )
							{
								if ( parenId < tp )
								{
									type.pop();
									lines.pop();
									parens.pop();
									columns.pop();
									lengths.pop();
								} else {
									break;
								}
							} else {
								break;
							}
							
						} while ( parens.count() );
						
						if ( parens.isEmpty() )
						{
							bMatch = true;
							m.type = PMatch::Mismatch;
							break;
						}
					}
				}
				
				if ( bInsert )
				{
					parens.push(parenId);
					columns.push(offset);
					lengths.push(length);
					type.push(true);
					lines.push(line);
				}
			} else if ( open ) {
				lines.push(line);
				type.push(false);
				columns.push(offset);
				parens.push(parenId);
				lengths.push(length);
			} else if ( close ) {
				if ( (bMatch = (parens.top() == parenId) ) )
				{
					bMatch &= parens.count() == 1;
					
					if ( bMatch )
					{
						m.type = PMatch::Match;
						
						break;
					} else if ( parens.count() ) {
						parens.pop();
						lines.pop();
						type.pop();
						lengths.pop();
						columns.pop();
					} else
						qWarning("bad paren nesting...");
				} else {
					bMatch = true;
					
					parens.pop();
					type.pop();
					m.line[0] = lines.pop();
					m.column[0] = columns.pop();
					m.length[0] = lengths.pop();
					
					m.type = PMatch::Mismatch;
					
					break;
				}
			}
		}
		
		if ( bMatch )
			break;
		}
		
		pos = 0;
		++line;
		
		QDocumentLine b = d->line(line);
		
		if ( !b.isValid() )
		{
			m.type = PMatch::Invalid;
			return;
		}
		
		m_tokens = b.tokens();
	}
	
	m.line[1] = line;
	m.column[1] = offset;
	m.length[1] = length;
}

void QNFADefinition::matchClose(QDocument *d, PMatch& m)
{
	int line = m.line[0];
	int pos = m.column[0];
	QVector<int> m_tokens = d->line(line).tokens();
	
	bool bMatch = false;
	int offset, length;
	QStack<int> lines;
	QStack<bool> type;
	QStack<int> parens;
	QStack<int> columns;
	QStack<int> lengths;
	
	forever
	{
		if ( m_tokens.count() > 3 )
		{
		
		int tokenIndex = m_tokens.first();
		
		while ( tokenIndex > 3 )
		{
			int rawId = m_tokens.at(--tokenIndex);
			length = m_tokens.at(--tokenIndex);
			offset = m_tokens.at(--tokenIndex);
			
			int tokenId = normalized(rawId);
			
			bool open = tokenId & QLanguageDefinition::OpenParen;
			bool close = tokenId & QLanguageDefinition::CloseParen;
			
			if ( !(open || close) || offset > pos )
				continue;
			
			int parenId = this->parenId(rawId);
			
			if ( open && close )
			{
				bool bInsert = true;
				
				if ( parens.count() )
				{
					if ( parens.top() == parenId )
					{
						//qDebug("submatch at %i", line);
						type.pop();
						lines.pop();
						parens.pop();
						columns.pop();
						lengths.pop();
						
						if ( parens.isEmpty() )
						{
							bMatch = true;
							//qDebug("match at %i", line);
							m.type = PMatch::Match;
							
							break;
						}
					} else {
						bInsert = false;
						
						do
						{
							int tp = parens.top();
							
							if ( type.top() )
							{
								if ( parenId < tp )
								{
									type.pop();
									lines.pop();
									parens.pop();
									columns.pop();
									lengths.pop();
								} else {
									break;
								}
							} else {
								break;
							}
							
						} while ( parens.count() );
						
						if ( parens.isEmpty() )
						{
							bMatch = true;
							m.type = PMatch::Mismatch;
							break;
						}
					}
				}
				
				if ( bInsert )
				{
					parens.push(parenId);
					columns.push(offset);
					lengths.push(length);
					type.push(true);
					lines.push(line);
				}
			} else if ( close ) {
				lines.push(line);
				type.push(false);
				columns.push(offset);
				parens.push(parenId);
				lengths.push(length);
			} else if ( open ) {
				if ( (bMatch = (parens.top() == parenId) ) )
				{
					bMatch &= parens.count() == 1;
					
					if ( bMatch )
					{
						m.type = PMatch::Match;
						
						break;
					} else if ( parens.count() ) {
						parens.pop();
						lines.pop();
						type.pop();
						lengths.pop();
						columns.pop();
					} else
						qWarning("bad paren nesting...");
				} else {
					bMatch = true;
					
					parens.pop();
					type.pop();
					m.line[0] = lines.pop();
					m.column[0] = columns.pop();
					m.length[0] = lengths.pop();
					
					m.type = PMatch::Mismatch;
					
					break;
				}
			}
		}
		
		if ( bMatch )
			break;
		
		}
		
		--line;
		
		QDocumentLine b = d->line(line);
		pos = b.length();
		
		if ( !b.isValid() )
		{
			m.type = PMatch::Invalid;
			return;
		}
		
		m_tokens = b.tokens();
	}
	
	m.line[1] = line;
	m.column[1] = offset;
	m.length[1] = length;
}

/*!
	\brief Return the indent to use when inserting a line at a given cursor position
*/
QString QNFADefinition::indent(const QDocumentCursor& c)
{
	if ( c.isNull() || c.line().isNull() )
		return QString();
	
	QDocumentLine b = c.line();
	int pos, max = qMin(c.columnNumber(), b.text().size());
	
	QString s = b.text().left(max);
	
	//qDebug("line %i, column %i : %s", b.lineNumber(), c.columnNumber(), qPrintable(s));
	
	for ( pos = 0; pos < max; pos++ )
		if ( !s.at(pos).isSpace() )
			break;
	
	int indent = 0;
	bool open = false;
	QString spaces = s.left(pos);
	
	#if 0
	foreach ( QParenthesis p, b.parentheses() )
	{
		if ( p.offset >= max )
			break;
		
		if ( !(p.role & QParenthesis::Indent) )
			continue;
		
		if ( p.role & QParenthesis::Open )
		{
			++indent;
		} else if ( p.role & QParenthesis::Close ) {
			--indent;
		}
	}
	#endif
	
	//qDebug("indent : %i", indent);
	
	if ( indent > 0 )
		spaces += QString(indent, '\t');
	
	return spaces;
}

/*!
	\brief Determines whether the given key event at the given position should cause unindent to happen
*/
bool QNFADefinition::unindent (const QDocumentCursor& c, const QString& ktxt)
{
	if ( c.isNull() || c.line().isNull() || ktxt.isEmpty() )
		return false;
	
	QDocumentLine b = c.line();
	int pos, max = qMin(c.columnNumber(), b.text().size());
	
	QString s = b.text();
	s.insert(max, ktxt);
	
	//qDebug("outdenting %s", qPrintable(s));
	for ( pos = 0; pos < max; ++pos ) 
		if ( !s.at(pos).isSpace() )
			break;
	
	if ( !pos || pos >= c.columnNumber() + ktxt.length() )
		return false;
	
	QString text = s.mid(pos);
	
	#if 0
	QNFAMatchContext cxt;
	QNFANotifier notify(text);
	
	QDocumentLine prev = c.document()->line(c.lineNumber() - 1);
	
	if ( prev.isValid() )
	{
		cxt = prev.matchContext();
	} else {
		// first line
		cxt = b.matchContext();
		cxt.reset();
	}
	
	::match(&cxt, text, &notify);
	
	const QVector<QParenthesis>& parens = notify.parentheses();
	
	if ( parens.isEmpty() )
		return false;
	
	//qDebug("%i parentheses", parens.count());
	QParenthesis p = parens.first();
	
	//qDebug("%i, %i, %i", p.offset, p.length, c.columnNumber());
	
	return (
				(p.role & QParenthesis::Close)
			&&
				(p.role & QParenthesis::Indent)
			&&
				(p.offset <= c.columnNumber() - pos)
			&&
				(p.offset + p.length > c.columnNumber() - pos)
			);
	#endif
	
	return false;
}

int QNFADefinition::findBlockEnd(QDocument *d, int line, bool *open)
{
	QDocumentLine b = d->line(line);
	
	QVector<int> pidx, m_tokens = b.tokens();
	
	if ( !b.isValid() || m_tokens.count() < 4 )
		return -1;
	
	int tokenIndex = 1;
	int tokenCount = m_tokens.first();
	
	while ( tokenIndex + 2 < tokenCount )
	{
		int offset = m_tokens.at(tokenIndex);
		int length = m_tokens.at(++tokenIndex);
		int rawId = m_tokens.at(++tokenIndex);
		++tokenIndex;
		
		int tokenId = normalized(rawId);
		
		bool popen = tokenId & QLanguageDefinition::OpenParen;
		bool pclose = tokenId & QLanguageDefinition::CloseParen;
		
		if ( !(popen || pclose) || !(tokenId & QLanguageDefinition::BlockFold) )
			continue;
		
		int parenId = this->parenId(rawId);
		
		if ( pclose && pidx.count() )
			pidx.pop_back();

		if ( popen )
			pidx << tokenIndex - 3;
		
	}
	
	if ( pidx.count() )
	{
		int firstPidx = pidx.first();
		
		PMatch m;
		m.line[0] = line;
		m.column[0] = m_tokens.at(firstPidx);
		m.length[0] = m_tokens.at(firstPidx + 1);
		
		matchOpen(d, m);
		
		if (
				(m.type == PMatch::Match)
			||
				(
					(m.type == PMatch::Mismatch)
				&&
					(m.line[0] == line)
				&&
					(m.column[0] == m_tokens.at(firstPidx))
				)
			)
		{
			if ( open )
			{
				int flags = blockFlags(d, m.line[1], 0);
				
				*open = QCE_FOLD_OPEN_COUNT(flags) | (m.type == PMatch::Mismatch);
			}
			
			return m.line[1];
		} else if ( m.type == PMatch::Invalid ) {
			return d->lines() - 1;
		}
	}
	
	#if 0
	QVector<QParenthesis> vp, lp = b.parentheses();
	
	while ( lp.count() )
	{
		QParenthesis p = lp.first();
		lp.pop_front();
		
		if ( !(p.role & QParenthesis::Fold) )
			continue;
		
		if ( p.role & QParenthesis::Close && vp.count() )
			vp.pop_back();

		if ( p.role & QParenthesis::Open )
			vp << p;
		
	}
	
	if ( vp.count() )
	{
		QParenthesis p = vp.first();
		
		PMatch m;
		m.line[0] = line;
		m.column[0] = p.offset;
		m.length[0] = p.length;
		
		matchOpen(d, m);
		
		if (
				(m.type == PMatch::Match)
			||
				(
					(m.type == PMatch::Mismatch)
				&&
					(m.line[0] == line)
				&&
					(m.column[0] == p.offset)
				)
			)
		{
			if ( open )
			{
				int flags = blockFlags(d, m.line[1], 0);

				*open = QCE_FOLD_OPEN_COUNT(flags) | (m.type == PMatch::Mismatch);
			}

			return m.line[1];
		} else if ( m.type == PMatch::Invalid ) {
			return d->lines() - 1;
		}
	}
	#endif
	
	return -1;
}


/*!
	\brief Expand a collapsed block at a given line
*/
void QNFADefinition::expand(QDocument *d, int line)
{
	QDocumentLine b = d->line(line);
	
	if ( !b.isValid() || !b.hasFlag(QDocumentLine::CollapsedBlockStart) )
		return;
	
	bool open = false;
	int end = findBlockEnd(d, line, &open);
	
	int depth = 1,
		count = end - line;
	
	if ( open )
		--count;
	
	if ( count <= 0 )
		return;
	
	b.setFlag(QDocumentLine::CollapsedBlockStart, false);
	
	bool shared = false;
	int i = line + 1;
	
	while ( i < end )
	{
		QDocumentLine l = d->line(i);
		l.setFlag(QDocumentLine::Hidden, false);
		//qDebug("+%i", i);
		if ( l.hasFlag(QDocumentLine::CollapsedBlockStart) )
		{
			bool bopen = false;
			int bend = findBlockEnd(d, i, &bopen);
			
			i = bend;

			if ( !bopen )
				++i;
			
			if ( bend == end )
			{
				shared = true;
			}
		} else {
			++i;
		}
	}
	
	if ( !shared )
	{
		d->line(end).setFlag(QDocumentLine::CollapsedBlockEnd, false);
	}
	
	if ( !open && !shared )
	{
		//qDebug("+%i", end);
		d->line(end).setFlag(QDocumentLine::Hidden, false);
	}
	
	//qDebug("expanding %i lines from %i", count, line);
	d->impl()->showEvent(line, count);
}

/*!
	\brief Collapse a text block at a given line
*/
void QNFADefinition::collapse(QDocument *d, int line)
{
	QDocumentLine b = d->line(line);
	
	if ( !b.isValid() || b.hasFlag(QDocumentLine::CollapsedBlockStart) )
		return;
	
	//qDebug("collapse line %i", line);
	
	bool open = false;
	int end = findBlockEnd(d, line, &open);
	
	int count = end - line,
		indent = 0;
	
	if ( open )
		--count;
	
	if ( count <= 0 )
		return;
	
	b.setFlag(QDocumentLine::CollapsedBlockStart);
	
	for ( int i = line + 1; i < end; ++i )
	{
		if ( !d->line(i).hasFlag(QDocumentLine::Hidden) )
		{
			//qDebug("-%i", i);
			d->line(i).setFlag(QDocumentLine::Hidden);
		}
	}
	
	d->line(end).setFlag(QDocumentLine::CollapsedBlockEnd);
	
	if ( !open )
	{
		if ( !d->line(end).hasFlag(QDocumentLine::Hidden) )
		{
			//qDebug("-%i", end);
			d->line(end).setFlag(QDocumentLine::Hidden);
		}
	}
	
	//qDebug("collapsing %i lines from %i", count, line);
	d->impl()->hideEvent(line, count);
	
	#if 0
	if ( !m_indentFold )
		return;
	
	int i = ln + 1, k;
	QDocumentLine block = d->line(i);
	
	if ( !block.isValid() )
		return;
	
	indent = block.firstChar();
	
	do
	{
		block = d->line(++i);
		k = block.firstChar();
	} while ( block.isValid() && ((k >= indent) || (k == -1)) );
	
	if ( k < indent )
	{
		block = d->line(--i);
		k = block.firstChar();
	}
	
	while ( block.isValid() && (i > ln) && (k == -1) )
	{
		block = d->line(--i);
		k = block.firstChar();
	}
	
	b.setFlag(QDocumentLine::CollapsedBlockStart);
	
	//qDebug("hidding from %i to %i", ln + 1, i);
	
	for ( int j = ln + 1; j <= i; ++j )
	{
		d->line(j).setFlag(QDocumentLine::Hidden);
	}
	
	d->line(i).setFlag(QDocumentLine::CollapsedBlockEnd);
	
	d->impl()->hideEvent(ln, i - ln);
	#endif
}

/*!
	\brief Compute the collapse state of a line
*/
int QNFADefinition::blockFlags(QDocument *d, int line, int depth) const
{
	QDocumentLine b = d->line(line);
	
	QVector<int> m_tokens = b.tokens();
	
	if ( !b.isValid() || m_tokens.count() < 4 )
		return None;
	
	int ret = None;
	short open = 0, close = 0;
	
	int tokenIndex = 1;
	int tokenCount = m_tokens.first();
	
	while ( tokenIndex + 2 < tokenCount )
	{
		int offset = m_tokens.at(tokenIndex);
		int length = m_tokens.at(++tokenIndex);
		int rawId = m_tokens.at(++tokenIndex);
		++tokenIndex;
		
		int tokenId = normalized(rawId);
		
		bool popen = tokenId & QLanguageDefinition::OpenParen;
		bool pclose = tokenId & QLanguageDefinition::CloseParen;
		
		if ( !(popen || pclose) || !(tokenId & QLanguageDefinition::BlockFold) )
			continue;
		
		int parenId = this->parenId(rawId);
		
		if ( pclose )
		{
			if ( open )
				--open;
			else
				++close;
			
		}
		
		if ( popen )
		{
			++open;
		}
	}
	
	#if 0
	foreach ( QParenthesis p, b.parentheses() )
	{
		if ( !(p.role & QParenthesis::Fold) )
			continue;
		
		if ( p.role & QParenthesis::Close )
		{
			if ( open )
				--open;
			else
				++close;
			
		}
		
		if ( p.role & QParenthesis::Open )
		{
			++open;
		}
	}
	#endif
	
	if ( b.hasFlag(QDocumentLine::CollapsedBlockStart) )
		ret = Collapsed;
	
	if ( open )
		ret |= Collapsible;
	
	if ( close )
		ret |= Closure;
	
	//qDebug("%i : ret : %i (%i, %i)", line, ret, ret & 0xffff0000, short(((short)0) | (ret & 0xffff)));
	
	return QCE_FOLD_FLAGS(ret, open, close);
}

void QNFADefinition::addContext(const QString& id, QNFA *nfa)
{
	//qDebug("registering context : %s", qPrintable(id));
	m_contexts[id] = nfa;
}

void QNFADefinition::flushEmbedRequests(const QString& lang)
{
	//qDebug("flushing requests...");
	QHash<QString, EmbedRequestList>::iterator it;
	
	it = m_pendingEmbeds.begin();
	
	while ( it != m_pendingEmbeds.end() )
	{
		QString r = it.key();
		
		if (
				r.startsWith(lang)
			&&
				(
					(r.count() == lang.count())
				||
					(r.at(lang.count() + 1) == ':')
				)
			)
		{
			QNFA *src = m_contexts.value(r);
			
			if ( !src )
			{
				++it;
				continue;
			}
			
			foreach( const EmbedRequest& request, *it )
			{
				//qDebug("embedding %s in 0x%x at index %i",
				//		qPrintable(r), request.target, request.index);
				
				embed(src, request.target, request.index);
			}
			
			it = m_pendingEmbeds.erase(it);
		} else {
			++it;
		}
	}
}

void QNFADefinition::addEmbedRequest(const QString& cxt, QNFA *dest)
{
	//qDebug("Adding request : %s", qPrintable(cxt));
	
	if ( m_contexts.contains(cxt) )
	{
		//qDebug("dealing with request : %s", qPrintable(cxt));
		embed(m_contexts[cxt], dest, dest->out.branch->count());
	} else {
		m_pendingEmbeds[cxt] << EmbedRequest(dest, dest->out.branch->count());
	}
}

void QNFADefinition::shareEmbedRequests(QNFA *src, QNFA *dest, int offset)
{
	QHash<QString, EmbedRequestList>::iterator it = m_pendingEmbeds.begin();
	
	while ( it != m_pendingEmbeds.end() )
	{
		foreach ( const EmbedRequest& request, *it )
		{
			if ( request.target == src )
			{
				//qDebug("sharing one embed request between 0x%x and 0x%x at %i + %i", src, dest, offset, request.index);
				it->append(EmbedRequest(dest, request.index + offset));
			}
		}
		
		++it;
	}
}
